#!/usr/bin/env python
# Safe Eyes is a utility to remind you to take break frequently
# to protect your eyes from eye strain.

# Copyright (C) 2016  Gobinath

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
This module creates the AboutDialog which shows the version and license.
"""

import os

from safeeyes import utility

ABOUT_DIALOG_GLADE = os.path.join(utility.BIN_DIRECTORY, "glade/about_dialog.glade")


class AboutDialog:
    """
    AboutDialog reads the about_dialog.glade and build the user interface using that file.
    It shows the application name with version, a small description, license and the GitHub url.
    """

    def __init__(self, version):
        builder = utility.create_gtk_builder(ABOUT_DIALOG_GLADE)
        builder.connect_signals(self)
        self.window = builder.get_object('window_about')
        builder.get_object('lbl_decription').set_label(_("Safe Eyes protects your eyes from eye strain (asthenopia) by reminding you to take breaks while you're working long hours at the computer"))
        builder.get_object('lbl_license').set_label(_('License') + ':')

        # Set the version at the runtime
        builder.get_object('lbl_app_name').set_label('Safe Eyes ' + version)

    def show(self):
        """
        Show the About dialog.
        """
        self.window.show_all()

    def on_window_delete(self, *args):
        """
        Window close event handler.
        """
        self.window.destroy()

    def on_close_clicked(self, *args):
        """
        Close button click event handler.
        """
        self.window.destroy()



#!/usr/bin/env python
# Safe Eyes is a utility to remind you to take break frequently
# to protect your eyes from eye strain.

# Copyright (C) 2017  Gobinath

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
This module contains the entity classes used by Safe Eyes and its plugins.
"""

import logging
import random
from enum import Enum
from dataclasses import dataclass

from packaging.version import parse

from safeeyes import utility


class Break:
    """
    An entity class which represents a break.
    """

    def __init__(self, break_type, name, time, duration, image, plugins):
        self.type = break_type
        self.name = name
        self.duration = duration
        self.image = image
        self.plugins = plugins
        self.time = time
        self.next = None

    def __str__(self):
        return 'Break: {{name: "{}", type: {}, duration: {}}}\n'.format(self.name, self.type, self.duration)

    def __repr__(self):
        return str(self)

    def is_long_break(self):
        """
        Check whether this break is a long break.
        """
        return self.type == BreakType.LONG_BREAK

    def is_short_break(self):
        """
        Check whether this break is a short break.
        """
        return self.type == BreakType.SHORT_BREAK

    def plugin_enabled(self, plugin_id, is_plugin_enabled):
        """
        Check whether this break supports the given plugin.
        """
        if self.plugins:
            return plugin_id in self.plugins
        else:
            return is_plugin_enabled


class BreakType(Enum):
    """
    Type of Safe Eyes breaks.
    """
    SHORT_BREAK = 1
    LONG_BREAK = 2


class BreakQueue:
    def __init__(self, config, context):
        self.context = context
        self.__current_break = None
        self.__current_long = 0
        self.__current_short = 0
        self.__short_break_time = config.get('short_break_interval')
        self.__long_break_time = config.get('long_break_interval')
        self.__is_random_order = config.get('random_order')
        self.__config = config

        self.__build_longs()
        self.__build_shorts()


        # Interface guarantees that short_interval >= 1
        # And that long_interval is a multiple of short_interval
        short_interval = config.get('short_break_interval')
        long_interval  = config.get('long_break_interval')
        self.__cycle_len = int(long_interval / short_interval)
        # To count every long break as a cycle in .next() if there are no short breaks
        if self.__short_queue is None:
            self.__cycle_len = 1

        # Restore the last break from session
        if not self.is_empty():
            last_break = context['session'].get('break')
            if last_break is not None:
                current_break = self.get_break()
                if last_break != current_break.name:
                    brk = self.next()
                    while brk != current_break and brk.name != last_break:
                        brk = self.next()

    def get_break(self, break_type = None):
        if self.__current_break is None:
            self.__current_break = self.next()

        if break_type is None or self.__current_break.type == break_type:
            return self.__current_break

        if break_type == BreakType.LONG_BREAK:
            if self.__long_queue is None:
                return None;
            return self.__long_queue[self.__current_long]

        if self.__short_queue is None:
            return None;
        return self.__short_queue[self.__current_short]

    def is_long_break(self):
        return self.__current_break is not None and self.__current_break.type == BreakType.LONG_BREAK

    def next(self, break_type = None):
        break_obj = None
        shorts = self.__short_queue
        longs  = self.__long_queue

        # Reset break that has just ended
        if self.is_long_break():
            self.__current_break.time = self.__long_break_time
            if self.__current_long == 0 and self.__is_random_order:
                # Shuffle queue
                self.__build_longs()
        elif self.__current_break:
            # Reduce the break time from the next long break (default)
            if longs:
                longs[self.__current_long].time -= shorts[self.__current_short].time
            if self.__current_short == 0 and self.__is_random_order:
                self.__build_shorts()

        if self.is_empty():
            return None

        if shorts is None:
            break_obj = self.__next_long()
        elif longs is None:
            break_obj = self.__next_short()
        elif break_type == BreakType.LONG_BREAK or longs[self.__current_long].time <= shorts[self.__current_short].time:
            break_obj = self.__next_long()
        else:
            break_obj = self.__next_short()

        self.__current_break = break_obj
        self.context['session']['break'] = self.__current_break.name

        return break_obj

    def reset(self):
        for break_object in self.__short_queue:
            break_object.time = self.__short_break_time

        for break_object in self.__long_queue:
            break_object.time = self.__long_break_time

    def is_empty(self, break_type = None):
        """
        Check if the given break type is empty or not. If the break_type is None, check for both short and long breaks.
        """
        if break_type == BreakType.SHORT_BREAK:
            return self.__short_queue is None
        elif break_type == BreakType.LONG_BREAK:
            return self.__long_queue is None
        else:
            return self.__short_queue is None and self.__long_queue is None

    def __next_short(self):
        longs  = self.__long_queue
        shorts = self.__short_queue
        break_obj = shorts[self.__current_short]
        self.context['break_type'] = 'short'

        # Update the index to next
        self.__current_short = (self.__current_short + 1) % len(shorts)

        return break_obj

    def __next_long(self):
        longs  = self.__long_queue
        break_obj = longs[self.__current_long]
        self.context['break_type'] = 'long'

        # Update the index to next
        self.__current_long = (self.__current_long + 1) % len(longs)

        return break_obj

    def __build_queue(self, break_type, break_configs, break_time, break_duration):
        """
        Build a queue of breaks.
        """
        size = len(break_configs)

        if 0 == size:
            # No breaks
            return None

        if self.__is_random_order:
            breaks_order = random.sample(break_configs, size)
        else:
            breaks_order = break_configs

        queue = [None] * size
        for i, break_config in enumerate(breaks_order):
            name = _(break_config['name'])
            duration = break_config.get('duration', break_duration)
            image = break_config.get('image')
            plugins = break_config.get('plugins', None)
            interval = break_config.get('interval', break_time)

            # Validate time value
            if not isinstance(duration, int) or duration <= 0:
                logging.error('Invalid break duration in: ' +
                              str(break_config))
                continue

            break_obj = Break(break_type, name, interval,
                              duration, image, plugins)
            queue[i] = break_obj

        return queue

    def __build_shorts(self):
        self.__short_queue = self.__build_queue(BreakType.SHORT_BREAK,
                                                  self.__config.get('short_breaks'),
                                                  self.__short_break_time,
                                                  self.__config.get('short_break_duration'))

    def __build_longs(self):
        self.__long_queue = self.__build_queue(BreakType.LONG_BREAK,
                                                 self.__config.get('long_breaks'),
                                                 self.__long_break_time,
                                                 self.__config.get('long_break_duration'))



class State(Enum):
    """
    Possible states of Safe Eyes.
    """
    START = 0,       # Starting scheduler
    WAITING = 1,     # User is working (waiting for next break)
    PRE_BREAK = 2,   # Preparing for break
    BREAK = 3,       # Break
    STOPPED = 4,     # Disabled
    QUIT = 5,        # Quitting
    RESTING = 6      # Resting (natural break)


class EventHook:
    """
    Hook to attach and detach listeners to system events.
    """

    def __init__(self):
        self.__handlers = []

    def __iadd__(self, handler):
        self.__handlers.append(handler)
        return self

    def __isub__(self, handler):
        self.__handlers.remove(handler)
        return self

    def fire(self, *args, **keywargs):
        """
        Fire all listeners attached with.
        """
        for handler in self.__handlers:
            if not handler(*args, **keywargs):
                return False
        return True


class Config:
    """
    The configuration of Safe Eyes.
    """

    def __init__(self, init=True):
        # Read the config files
        self.__user_config = utility.load_json(utility.CONFIG_FILE_PATH)
        self.__system_config = utility.load_json(
            utility.SYSTEM_CONFIG_FILE_PATH)
        # If there any breaking changes in long_breaks, short_breaks or any other keys, use the __force_upgrade list
        self.__force_upgrade = []
        # self.__force_upgrade = ['long_breaks', 'short_breaks']

        if init:
            # if create_startup_entry finds a broken autostart symlink, it will repair it
            utility.create_startup_entry(force=False)
            if self.__user_config is None:
                utility.initialize_safeeyes()
                self.__user_config = self.__system_config
                self.save()
            else:
                system_config_version = self.__system_config['meta']['config_version']
                meta_obj = self.__user_config.get('meta', None)
                if meta_obj is None:
                    # Corrupted user config
                    self.__user_config = self.__system_config
                else:
                    user_config_version = str(
                        meta_obj.get('config_version', '0.0.0'))
                    if parse(user_config_version) != parse(system_config_version):
                        # Update the user config
                        self.__merge_dictionary(
                            self.__user_config, self.__system_config)
                        self.__user_config = self.__system_config
                        # Update the style sheet
                        utility.replace_style_sheet()

            utility.merge_plugins(self.__user_config)
            self.save()

    def __merge_dictionary(self, old_dict, new_dict):
        """
        Merge the dictionaries.
        """
        for key in new_dict:
            if key == "meta" or key in self.__force_upgrade:
                continue
            if key in old_dict:
                new_value = new_dict[key]
                old_value = old_dict[key]
                if type(new_value) is type(old_value):
                    # Both properties have same type
                    if isinstance(new_value, dict):
                        self.__merge_dictionary(old_value, new_value)
                    else:
                        new_dict[key] = old_value

    def clone(self):
        config = Config(init=False)
        return config

    def save(self):
        """
        Save the configuration to file.
        """
        utility.write_json(utility.CONFIG_FILE_PATH, self.__user_config)

    def get(self, key, default_value=None):
        """
        Get the value.
        """
        value = self.__user_config.get(key, default_value)
        if value is None:
            value = self.__system_config.get(key, None)
        return value

    def set(self, key, value):
        """
        Set the value.
        """
        self.__user_config[key] = value

    def __eq__(self, config):
        return self.__user_config == config.__user_config

    def __ne__(self, config):
        return self.__user_config != config.__user_config


class TrayAction:
    """
    Data object wrapping name, icon and action.
    """

    def __init__(self, name, icon, action, system_icon):
        self.name = name
        self.__icon = icon
        self.action = action
        self.system_icon = system_icon
        self.__toolbar_buttons = []

    def get_icon(self):
        if self.system_icon:
            return self.__icon
        else:
            image = utility.load_and_scale_image(self.__icon, 16, 16)
            image.show()
            return image

    def add_toolbar_button(self, button):
        self.__toolbar_buttons.append(button)

    def reset(self):
        for button in self.__toolbar_buttons:
            button.hide()
        self.__toolbar_buttons.clear()

    @classmethod
    def build(cls, name, icon_path, icon_id, action):
        image = utility.load_and_scale_image(icon_path, 12, 12)
        if image is None:
            return TrayAction(name, icon_id, action, True)
        else:
            return TrayAction(name, icon_path, action, False)

@dataclass
class PluginDependency:
    message: str
    link: str|None = None

class RequiredPluginException(Exception):
    def __init__(self, plugin_id, plugin_name: str, message: str|PluginDependency):
        if isinstance(message, PluginDependency):
            msg = message.message
        else:
            msg = message

        super().__init__(msg)

        self.plugin_id = plugin_id
        self.plugin_name = plugin_name
        self.message = message

    def get_plugin_id(self):
        return self.plugin_id

    def get_plugin_name(self):
        return self.plugin_name

    def get_message(self):
        return self.message



#!/usr/bin/env python3
# Safe Eyes is a utility to remind you to take break frequently
# to protect your eyes from eye strain.

# Copyright (C) 2016  Gobinath

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
Safe Eyes is a utility to remind you to take break frequently to protect your eyes from eye strain.
"""
import argparse
import gettext
import locale
import logging
import signal
import sys

import psutil
from safeeyes import utility
from safeeyes.model import Config
from safeeyes.safeeyes import SafeEyes
from safeeyes.safeeyes import SAFE_EYES_VERSION
from safeeyes.rpc import RPCClient

gettext.install('safeeyes', utility.LOCALE_PATH)


def __running():
    """
    Check if SafeEyes is already running.
    """
    process_count = 0
    for proc in psutil.process_iter():
        if not proc.cmdline:
            continue
        try:
            # Check if safeeyes is in process arguments
            if callable(proc.cmdline):
                # Latest psutil has cmdline function
                cmd_line = proc.cmdline()
            else:
                # In older versions cmdline was a list object
                cmd_line = proc.cmdline
            if ('python3' in cmd_line[0] or 'python' in cmd_line[0]) and ('safeeyes' in cmd_line[1] or 'safeeyes' in cmd_line):
                process_count += 1
                if process_count > 1:
                    return True

        # Ignore if process does not exist or does not have command line args
        except (IndexError, psutil.NoSuchProcess):
            pass
    return False


def main():
    """
    Start the Safe Eyes.
    """
    system_locale = gettext.translation('safeeyes', localedir=utility.LOCALE_PATH, languages=[utility.system_locale(), 'en_US'], fallback=True)
    system_locale.install()
    try:
        # locale.bindtextdomain is required for Glade files
        locale.bindtextdomain('safeeyes', utility.LOCALE_PATH)
    except AttributeError:
        logging.warning('installed python\'s gettext module does not support locale.bindtextdomain. locale.bindtextdomain is required for Glade files')


    parser = argparse.ArgumentParser(prog='safeeyes')
    group = parser.add_mutually_exclusive_group()
    group.add_argument('-a', '--about', help=_('show the about dialog'), action='store_true')
    group.add_argument('-d', '--disable', help=_('disable the currently running safeeyes instance'), action='store_true')
    group.add_argument('-e', '--enable', help=_('enable the currently running safeeyes instance'), action='store_true')
    group.add_argument('-q', '--quit', help=_('quit the running safeeyes instance and exit'), action='store_true')
    group.add_argument('-s', '--settings', help=_('show the settings dialog'), action='store_true')
    group.add_argument('-t', '--take-break', help=_('Take a break now').lower(), action='store_true')
    parser.add_argument('--debug', help=_('start safeeyes in debug mode'), action='store_true')
    parser.add_argument('--status', help=_('print the status of running safeeyes instance and exit'), action='store_true')
    parser.add_argument('--version', action='version', version='%(prog)s ' + SAFE_EYES_VERSION)
    args = parser.parse_args()

    # Initialize the logging
    utility.initialize_logging(args.debug)
    utility.initialize_platform()
    config = Config()
    utility.create_user_stylesheet_if_missing()

    if __running():
        logging.info("Safe Eyes is already running")
        if not config.get("use_rpc_server", True):
            # RPC sever is disabled
            print(_('Safe Eyes is running without an RPC server. Turn it on to use command-line arguments.'))
            sys.exit(0)
            return
        rpc_client = RPCClient(config.get('rpc_port'))
        if args.about:
            rpc_client.show_about()
        elif args.disable:
            rpc_client.disable_safeeyes()
        elif args.enable:
            rpc_client.enable_safeeyes()
        elif args.settings:
            rpc_client.show_settings()
        elif args.take_break:
            rpc_client.take_break()
        elif args.status:
            print(rpc_client.status())
        elif args.quit:
            rpc_client.quit()
        else:
            # Default behavior is opening settings
            rpc_client.show_settings()
        sys.exit(0)
    else:
        if args.status:
            print(_('Safe Eyes is not running'))
            sys.exit(0)
        elif not args.quit:
            logging.info("Starting Safe Eyes")
            safe_eyes = SafeEyes(system_locale, config, args)
            safe_eyes.start()


if __name__ == '__main__':
    signal.signal(signal.SIGINT, signal.SIG_DFL)    # Handle Ctrl + C
    main()



#!/usr/bin/env python
# Safe Eyes is a utility to remind you to take break frequently
# to protect your eyes from eye strain.

# Copyright (C) 2017  Gobinath

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
SafeEyes connects all the individual components and provide the complete application.
"""

import atexit
import logging
import os
from threading import Timer

import gi
from safeeyes import utility
from safeeyes.ui.about_dialog import AboutDialog
from safeeyes.ui.break_screen import BreakScreen
from safeeyes.ui.required_plugin_dialog import RequiredPluginDialog
from safeeyes.model import State, RequiredPluginException
from safeeyes.rpc import RPCServer
from safeeyes.plugin_manager import PluginManager
from safeeyes.core import SafeEyesCore
from safeeyes.ui.settings_dialog import SettingsDialog

gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, Gio

SAFE_EYES_VERSION = "2.2.2"


class SafeEyes(Gtk.Application):
    """
    This class represents a runnable Safe Eyes instance.
    """

    required_plugin_dialog_active = False

    def __init__(self, system_locale, config, cli_args):
        super().__init__(
            application_id="io.github.slgobinath.SafeEyes",
            flags=Gio.ApplicationFlags.FLAGS_NONE ## This is necessary for compatibility with Ubuntu 22.04.
        )
        self.active = False
        self.break_screen = None
        self.safe_eyes_core = None
        self.config = config
        self.context = {}
        self.plugins_manager = None
        self.settings_dialog_active = False
        self.rpc_server = None
        self._status = ''
        self.cli_args = cli_args
        self.system_locale = system_locale

    def start(self):
        """
        Start Safe Eyes
        """
        self.run()

    def do_startup(self):
        Gtk.Application.do_startup(self)

        logging.info('Starting up Application')

        # Initialize the Safe Eyes Context
        self.context['version'] = SAFE_EYES_VERSION
        self.context['desktop'] = utility.desktop_environment()
        self.context['is_wayland'] = utility.is_wayland()
        self.context['locale'] = self.system_locale
        self.context['api'] = {}
        self.context['api']['show_settings'] = lambda: utility.execute_main_thread(
            self.show_settings)
        self.context['api']['show_about'] = lambda: utility.execute_main_thread(
            self.show_about)
        self.context['api']['enable_safeeyes'] = lambda next_break_time=-1, reset_breaks=False: \
            utility.execute_main_thread(self.enable_safeeyes, next_break_time, reset_breaks)
        self.context['api']['disable_safeeyes'] = lambda status=None, is_resting=False: utility.execute_main_thread(
            self.disable_safeeyes, status, is_resting)
        self.context['api']['status'] = self.status
        self.context['api']['quit'] = lambda: utility.execute_main_thread(
            self.quit)
        if self.config.get('persist_state'):
            self.context['session'] = utility.open_session()
        else:
            self.context['session'] = {'plugin': {}}

        self.break_screen = BreakScreen(
            self.context, self.on_skipped, self.on_postponed, utility.STYLE_SHEET_PATH)
        self.break_screen.initialize(self.config)
        self.plugins_manager = PluginManager()
        self.safe_eyes_core = SafeEyesCore(self.context)
        self.safe_eyes_core.on_pre_break += self.plugins_manager.pre_break
        self.safe_eyes_core.on_start_break += self.on_start_break
        self.safe_eyes_core.start_break += self.start_break
        self.safe_eyes_core.on_count_down += self.countdown
        self.safe_eyes_core.on_stop_break += self.stop_break
        self.safe_eyes_core.on_update_next_break += self.update_next_break
        self.safe_eyes_core.initialize(self.config)
        self.context['api']['take_break'] = self.take_break
        self.context['api']['has_breaks'] = self.safe_eyes_core.has_breaks
        self.context['api']['postpone'] = self.safe_eyes_core.postpone
        self.context['api']['get_break_time'] = self.safe_eyes_core.get_break_time

        try:
            self.plugins_manager.init(self.context, self.config)
        except RequiredPluginException as e:
            self.show_required_plugin_dialog(e.get_plugin_id(), e.get_plugin_name(), e.get_message())
            self.required_plugin_dialog_active = True

        self.hold()

        atexit.register(self.persist_session)

        if self.config.get('use_rpc_server', True):
            self.__start_rpc_server()

        if not self.required_plugin_dialog_active and self.safe_eyes_core.has_breaks():
            self.active = True
            self.context['state'] = State.START
            self.plugins_manager.start()		# Call the start method of all plugins
            self.safe_eyes_core.start()
            self.handle_system_suspend()

    def do_activate(self):
        logging.info('Application activated')

        if self.cli_args.about:
            self.show_about()
        elif self.cli_args.disable:
            self.disable_safeeyes()
        elif self.cli_args.enable:
            self.enable_safeeyes()
        elif self.cli_args.settings:
            self.show_settings()
        elif self.cli_args.take_break:
            self.take_break()

    def show_settings(self):
        """
        Listen to tray icon Settings action and send the signal to Settings dialog.
        """
        if not self.settings_dialog_active:
            logging.info("Show Settings dialog")
            self.settings_dialog_active = True
            settings_dialog = SettingsDialog(
                self.config.clone(), self.save_settings)
            settings_dialog.show()

    def show_required_plugin_dialog(self, plugin_id, plugin_name, message):
        logging.info("Show RequiredPlugin dialog")
        dialog = RequiredPluginDialog(
            plugin_id,
            plugin_name,
            message,
            self.quit,
            lambda: self.disable_plugin(plugin_id)
        )
        dialog.show()

    def disable_plugin(self, plugin_id):
        """
        Temporarily disable plugin, and restart SafeEyes.
        """
        config = self.config.clone()

        for plugin in config.get('plugins'):
            if plugin['id'] == plugin_id:
                plugin['enabled'] = False

        self.required_plugin_dialog_active = False

        self.restart(config, set_active=True)

    def show_about(self):
        """
        Listen to tray icon About action and send the signal to About dialog.
        """
        logging.info("Show About dialog")
        about_dialog = AboutDialog(SAFE_EYES_VERSION)
        about_dialog.show()

    def quit(self):
        """
        Listen to the tray menu quit action and stop the core, notification and the app itself.
        """
        logging.info("Quit Safe Eyes")
        self.break_screen.close()
        self.context['state'] = State.QUIT
        self.plugins_manager.stop()
        self.safe_eyes_core.stop()
        self.plugins_manager.exit()
        self.__stop_rpc_server()
        self.persist_session()

        super().quit()

    def handle_suspend_callback(self, sleeping):
        """
        If the system goes to sleep, Safe Eyes stop the core if it is already active.
        If it was active, Safe Eyes will become active after wake up.
        """
        if sleeping:
            # Sleeping / suspending
            if self.active:
                logging.info("Stop Safe Eyes due to system suspend")
                self.plugins_manager.stop()
                self.safe_eyes_core.stop(True)
        else:
            # Resume from sleep
            if self.active and self.safe_eyes_core.has_breaks():
                logging.info("Resume Safe Eyes after system wakeup")
                self.plugins_manager.start()
                self.safe_eyes_core.start()

    def handle_suspend_signal(self, proxy, sender, signal, parameters):
        if signal != "PrepareForSleep":
            return

        (sleeping, ) = parameters

        self.handle_suspend_callback(sleeping)

    def handle_system_suspend(self):
        """
        Setup system suspend listener.
        """
        self.suspend_proxy = Gio.DBusProxy.new_for_bus_sync(
            bus_type=Gio.BusType.SYSTEM,
            flags=Gio.DBusProxyFlags.DO_NOT_LOAD_PROPERTIES,
            info=None,
            name='org.freedesktop.login1',
            object_path='/org/freedesktop/login1',
            interface_name='org.freedesktop.login1.Manager',
            cancellable=None,
        )
        self.suspend_proxy.connect('g-signal', self.handle_suspend_signal)

    def on_skipped(self):
        """
        Listen to break screen Skip action and send the signal to core.
        """
        logging.info("User skipped the break")
        self.safe_eyes_core.skip()
        self.plugins_manager.stop_break()

    def on_postponed(self):
        """
        Listen to break screen Postpone action and send the signal to core.
        """
        logging.info("User postponed the break")
        self.safe_eyes_core.postpone()
        self.plugins_manager.stop_break()

    def save_settings(self, config):
        """
        Listen to Settings dialog Save action and write to the config file.
        """
        self.settings_dialog_active = False

        if self.config == config:
            # Config is not modified
            return

        logging.info("Saving settings to safeeyes.json")
        # Stop the Safe Eyes core
        if self.active:
            self.plugins_manager.stop()
            self.safe_eyes_core.stop()

        # Write the configuration to file
        config.save()
        self.persist_session()

        self.restart(config)

    def restart(self, config, set_active=False):
        logging.info("Initialize SafeEyesCore with modified settings")

        if self.rpc_server is None and config.get('use_rpc_server'):
            # RPC server wasn't running but now enabled
            self.__start_rpc_server()
        elif self.rpc_server is not None and not config.get('use_rpc_server'):
            # RPC server was running but now disabled
            self.__stop_rpc_server()

        # Restart the core and initialize the components
        self.config = config
        self.safe_eyes_core.initialize(config)
        self.break_screen.initialize(config)

        try:
            self.plugins_manager.init(self.context, self.config)
        except RequiredPluginException as e:
            self.show_required_plugin_dialog(e.get_plugin_id(), e.get_plugin_name(), e.get_message())
            self.required_plugin_dialog_active = True
            return

        if set_active:
            self.active = True

        if self.active and self.safe_eyes_core.has_breaks():
            # 1 sec delay is required to give enough time for core to be stopped
            Timer(1.0, self.safe_eyes_core.start).start()
            self.plugins_manager.start()

    def enable_safeeyes(self, scheduled_next_break_time=-1, reset_breaks=False):
        """
        Listen to tray icon enable action and send the signal to core.
        """
        if not self.required_plugin_dialog_active and not self.active and self.safe_eyes_core.has_breaks():
            self.active = True
            self.safe_eyes_core.start(scheduled_next_break_time, reset_breaks)
            self.plugins_manager.start()

    def disable_safeeyes(self, status=None, is_resting = False):
        """
        Listen to tray icon disable action and send the signal to core.
        """
        if self.active:
            self.active = False
            self.plugins_manager.stop()
            self.safe_eyes_core.stop(is_resting)
            if status is None:
                status = _('Disabled until restart')
            self._status = status

    def on_start_break(self, break_obj):
        """
        Pass the break information to plugins.
        """
        if not self.plugins_manager.start_break(break_obj):
            return False
        return True

    def start_break(self, break_obj):
        """
        Pass the break information to break screen.
        """
        # Get the HTML widgets content from plugins
        widget = self.plugins_manager.get_break_screen_widgets(break_obj)
        actions = self.plugins_manager.get_break_screen_tray_actions(break_obj)
        self.break_screen.show_message(break_obj, widget, actions)

    def countdown(self, countdown, seconds):
        """
        Pass the countdown to plugins and break screen.
        """
        self.break_screen.show_count_down(countdown, seconds)
        self.plugins_manager.countdown(countdown, seconds)
        return True

    def update_next_break(self, break_obj, break_time):
        """
        Update the next break to plugins and save the session.
        """
        self.plugins_manager.update_next_break(break_obj, break_time)
        self._status = _('Next break at %s') % (
            utility.format_time(break_time))
        if self.config.get('persist_state'):
            utility.write_json(utility.SESSION_FILE_PATH,
                               self.context['session'])

    def stop_break(self):
        """
        Stop the current break.
        """
        self.break_screen.close()
        self.plugins_manager.stop_break()
        return True

    def take_break(self, break_type = None):
        """
        Take a break now.
        """
        utility.execute_main_thread(self.safe_eyes_core.take_break, break_type)

    def status(self):
        """
        Return the status of Safe Eyes.
        """
        return self._status

    def persist_session(self):
        """
        Save the session object to the session file.
        """
        if self.config.get('persist_state'):
            utility.write_json(utility.SESSION_FILE_PATH,
                               self.context['session'])
        else:
            utility.delete(utility.SESSION_FILE_PATH)

    def __start_rpc_server(self):
        if self.rpc_server is None:
            self.rpc_server = RPCServer(self.config.get('rpc_port'), self.context)
            self.rpc_server.start()

    def __stop_rpc_server(self):
        if self.rpc_server is not None:
            self.rpc_server.stop()
            self.rpc_server = None



# Safe Eyes is a utility to remind you to take break frequently
# to protect your eyes from eye strain.

# Copyright (C) 2017  Gobinath

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from safeeyes import utility
from safeeyes.model import PluginDependency

import gi
gi.require_version('Gio', '2.0')
from gi.repository import Gio

def validate(plugin_config, plugin_settings):
    dbus_proxy = Gio.DBusProxy.new_for_bus_sync(
        bus_type=Gio.BusType.SESSION,
        flags=Gio.DBusProxyFlags.DO_NOT_LOAD_PROPERTIES,
        info=None,
        name='org.freedesktop.DBus',
        object_path='/org/freedesktop/DBus',
        interface_name='org.freedesktop.DBus',
        cancellable=None,
    )

    if dbus_proxy.NameHasOwner('(s)', 'org.kde.StatusNotifierWatcher'):
        return None
    else:
        return PluginDependency(
            message=_("Please install service providing tray icons for your desktop environment."),
            link="https://github.com/slgobinath/SafeEyes/wiki/How-to-install-backend-for-Safe-Eyes-tray-icon"
        )

    command = None
    if utility.IS_WAYLAND:
        if utility.DESKTOP_ENVIRONMENT == "gnome":
            return None
        command = "wlrctl"
    else:
        command = "xprop"
    if not utility.command_exist(command):
        return _("Please install the command-line tool '%s'") % command
    else:
        return None



# Safe Eyes is a utility to remind you to take break frequently
# to protect your eyes from eye strain.

# Copyright (C) 2017  Gobinath

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import datetime
from safeeyes.model import BreakType
import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gio, GLib
import logging
from safeeyes import utility
import threading
import time

"""
Safe Eyes tray icon plugin
"""

context = None
tray_icon = None
safeeyes_config = None

SNI_NODE_INFO = Gio.DBusNodeInfo.new_for_xml("""
<?xml version="1.0" encoding="UTF-8"?>
<node>
    <interface name="org.kde.StatusNotifierItem">
        <property name="Category" type="s" access="read"/>
        <property name="Id" type="s" access="read"/>
        <property name="Title" type="s" access="read"/>
        <property name="ToolTip" type="(sa(iiay)ss)" access="read"/>
        <property name="Menu" type="o" access="read"/>
        <property name="ItemIsMenu" type="b" access="read"/>
        <property name="IconName" type="s" access="read"/>
        <property name="IconThemePath" type="s" access="read"/>
        <property name="Status" type="s" access="read"/>
        <signal name="NewIcon"/>
        <signal name="NewTooltip"/>
    </interface>
</node>""").interfaces[0]

MENU_NODE_INFO = Gio.DBusNodeInfo.new_for_xml("""
<?xml version="1.0" encoding="UTF-8"?>
<node>
    <interface name="com.canonical.dbusmenu">
        <method name="GetLayout">
            <arg type="i" direction="in"/>
            <arg type="i" direction="in"/>
            <arg type="as" direction="in"/>
            <arg type="u" direction="out"/>
            <arg type="(ia{sv}av)" direction="out"/>
        </method>
        <method name="GetGroupProperties">
			<arg type="ai" name="ids" direction="in"/>
			<arg type="as" name="propertyNames" direction="in" />
			<arg type="a(ia{sv})" name="properties" direction="out" />
		</method>
        <method name="GetProperty">
			<arg type="i" name="id" direction="in"/>
			<arg type="s" name="name" direction="in"/>
			<arg type="v" name="value" direction="out"/>
		</method>
        <method name="Event">
            <arg type="i" direction="in"/>
            <arg type="s" direction="in"/>
            <arg type="v" direction="in"/>
            <arg type="u" direction="in"/>
        </method>
        <method name="EventGroup">
			<arg type="a(isvu)" name="events" direction="in" />
			<arg type="ai" name="idErrors" direction="out" />
		</method>
        <method name="AboutToShow">
            <arg type="i" direction="in"/>
            <arg type="b" direction="out"/>
        </method>
        <method name="AboutToShowGroup">
			<arg type="ai" name="ids" direction="in" />
			<arg type="ai" name="updatesNeeded" direction="out" />
			<arg type="ai" name="idErrors" direction="out" />
		</method>
        <signal name="LayoutUpdated">
            <arg type="u"/>
            <arg type="i"/>
        </signal>
    </interface>
</node>""").interfaces[0]

class DBusService:
    def __init__(self, interface_info, object_path, bus):
        self.interface_info = interface_info
        self.bus = bus
        self.object_path = object_path
        self.registration_id = None

    def register(self):
        self.registration_id = self.bus.register_object(
            object_path=self.object_path,
            interface_info=self.interface_info,
            method_call_closure=self.on_method_call,
            get_property_closure=self.on_get_property
        )

        if not self.registration_id:
            raise GLib.Error(f"Failed to register object with path {self.object_path}")

        self.interface_info.cache_build()

    def unregister(self):
        self.interface_info.cache_release()

        if self.registration_id is not None:
            self.bus.unregister_object(self.registration_id)
            self.registration_id = None

    def on_method_call(self, _connection, _sender, _path, _interface_name, method_name, parameters, invocation):
        method_info = self.interface_info.lookup_method(method_name)
        method = getattr(self, method_name)
        result = method(*parameters.unpack())
        out_arg_types = "".join([arg.signature for arg in method_info.out_args])
        return_value = None

        if method_info.out_args:
            return_value = GLib.Variant(f"({out_arg_types})", result)

        invocation.return_value(return_value)

    def on_get_property(self, _connection, _sender, _path, _interface_name, property_name):
        property_info = self.interface_info.lookup_property(property_name)
        return GLib.Variant(property_info.signature, getattr(self, property_name))

    def emit_signal(self, signal_name, args = None):
        signal_info = self.interface_info.lookup_signal(signal_name)
        if len(signal_info.args) == 0:
            parameters = None
        else:
            arg_types = "".join([arg.signature for arg in signal_info.args])
            parameters = GLib.Variant(f"({arg_types})", args)
        self.bus.emit_signal(
            destination_bus_name=None,
            object_path=self.object_path,
            interface_name=self.interface_info.name,
            signal_name=signal_name,
            parameters=parameters
        )

class DBusMenuService(DBusService):
    DBUS_SERVICE_PATH = '/io/github/slgobinath/SafeEyes/Menu'

    revision = 0

    items = []
    idToItems = {}

    def __init__(self, session_bus, context, items):
        super().__init__(
            interface_info=MENU_NODE_INFO,
            object_path=self.DBUS_SERVICE_PATH,
            bus=session_bus
        )

        self.set_items(items)

    def set_items(self, items):
        self.items = items

        self.idToItems = self.getItemsFlat(items, {})

        self.revision += 1

        self.LayoutUpdated(self.revision, 0)

    @staticmethod
    def getItemsFlat(items, idToItems):
        for item in items:
            if item.get('hidden', False) == True:
                continue

            idToItems[item['id']] = item

            if 'children' in item:
                idToItems = DBusMenuService.getItemsFlat(item['children'], idToItems)

        return idToItems

    @staticmethod
    def singleItemToDbus(item):
        props = DBusMenuService.itemPropsToDbus(item)

        return (item['id'], props)

    @staticmethod
    def itemPropsToDbus(item):
        result = {}

        string_props = ['label', 'icon-name', 'type', 'children-display']
        for key in string_props:
            if key in item:
                result[key] = GLib.Variant('s', item[key])

        bool_props = ['enabled']
        for key in bool_props:
            if key in item:
                result[key] = GLib.Variant('b', item[key])

        return result

    @staticmethod
    def itemToDbus(item, recursion_depth):
        if item.get('hidden', False) == True:
            return None

        props = DBusMenuService.itemPropsToDbus(item)

        children = []
        if recursion_depth > 1 or recursion_depth == -1:
            if "children" in item:
                children = [DBusMenuService.itemToDbus(item, recursion_depth - 1) for item in item['children']]
                children = [i for i in children if i is not None]

        return GLib.Variant("(ia{sv}av)", (item['id'], props, children))

    def findItemsWithParent(self, parent_id, items):
        for item in items:
            if item.get('hidden', False) == True:
                continue
            if 'children' in item:
                if item['id'] == parent_id:
                    return item['children']
                else:
                    ret = self.findItemsWithParent(parent_id, item['children'])
                    if ret is not None:
                        return ret
        return None

    def GetLayout(self, parent_id, recursion_depth, property_names):
        children = []

        if parent_id == 0:
            children = self.items
        else:
            children = self.findItemsWithParent(parent_id, self.items)
            if children is None:
                children = []

        children = [self.itemToDbus(item, recursion_depth) for item in children]
        children = [i for i in children if i is not None]

        ret = (
            self.revision,
            (
                0,
                { 'children-display': GLib.Variant('s', 'submenu') },
                children
            )
        )

        return ret

    def GetGroupProperties(self, ids, property_names):
        ret = []

        for idx in ids:
            if idx in self.idToItems:
                props = DBusMenuService.singleItemToDbus(self.idToItems[idx])
                if props is not None:
                    ret.append(props)

        return (ret,)

    def GetProperty(self, idx, name):
        ret = None

        if idx in self.idToItems:
            props = DBusMenuService.singleItemToDbus(self.idToItems[idx])
            if props is not None and name in props:
                ret = props[name]

        return ret

    def Event(self, idx, event_id, data, timestamp):
        if event_id != "clicked":
            return

        if idx in self.idToItems:
            item = self.idToItems[idx]
            if 'callback' in item:
                item['callback']()

    def EventGroup(self, events):
        not_found = []

        for (idx, event_id, data, timestamp) in events:
            if idx not in self.idToItems:
                not_found.append(idx)
                continue

            if event_id != "clicked":
                continue

            item = self.idToItems[idx]
            if 'callback' in item:
                item['callback']()

        return not_found

    def AboutToShow(self, item_id):
        return (False,)

    def AboutToShowGroup(self, ids):
        not_found = []

        for idx in ids:
            if idx not in self.idToItems:
                not_found.append(idx)
                continue

        return ([], not_found)

    def LayoutUpdated(self, revision, parent):
        self.emit_signal(
            'LayoutUpdated',
            (revision, parent)
        )

class StatusNotifierItemService(DBusService):
    DBUS_SERVICE_PATH = '/org/ayatana/NotificationItem/io_github_slgobinath_SafeEyes'

    Category = 'ApplicationStatus'
    Id = 'io.github.slgobinath.SafeEyes'
    Title = 'Safe Eyes'
    Status = 'Active'
    IconName = 'io.github.slgobinath.SafeEyes-enabled'
    IconThemePath = ''
    ToolTip = ('', [], 'Safe Eyes', '')
    ItemIsMenu = True
    Menu = None

    def __init__(self, session_bus, context, menu_items):
        super().__init__(
            interface_info=SNI_NODE_INFO,
            object_path=self.DBUS_SERVICE_PATH,
            bus=session_bus
        )

        self.bus = session_bus

        self._menu = DBusMenuService(session_bus, context, menu_items)
        self.Menu = self._menu.DBUS_SERVICE_PATH

    def register(self):
        self._menu.register()
        super().register()

        watcher = Gio.DBusProxy.new_sync(
            connection=self.bus,
            flags=Gio.DBusProxyFlags.DO_NOT_LOAD_PROPERTIES,
            info=None,
            name='org.kde.StatusNotifierWatcher',
            object_path='/StatusNotifierWatcher',
            interface_name='org.kde.StatusNotifierWatcher',
            cancellable=None,
        )

        watcher.RegisterStatusNotifierItem('(s)', self.DBUS_SERVICE_PATH)

    def unregister(self):
        super().unregister()
        self._menu.unregister()

    def set_items(self, items):
        self._menu.set_items(items)

    def set_icon(self, icon):
        self.IconName = icon

        self.emit_signal(
            'NewIcon'
        )

    def set_tooltip(self, title, description):
        self.ToolTip = ('', [], title, description)

        self.emit_signal(
            'NewTooltip'
        )

class TrayIcon:
    """
    Create and show the tray icon along with the tray menu.
    """

    def __init__(self, context, plugin_config):
        self.context = context
        self.on_show_settings = context['api']['show_settings']
        self.on_show_about = context['api']['show_about']
        self.quit = context['api']['quit']
        self.enable_safeeyes = context['api']['enable_safeeyes']
        self.disable_safeeyes = context['api']['disable_safeeyes']
        self.take_break = context['api']['take_break']
        self.has_breaks = context['api']['has_breaks']
        self.get_break_time = context['api']['get_break_time']
        self.plugin_config = plugin_config
        self.date_time = None
        self.active = True
        self.wakeup_time = None
        self.idle_condition = threading.Condition()
        self.lock = threading.Lock()
        self.allow_disabling = plugin_config['allow_disabling']
        self.animate = False
        self.menu_locked = False

        session_bus = Gio.bus_get_sync(Gio.BusType.SESSION)

        self.sni_service = StatusNotifierItemService(
            session_bus,
            context,
            menu_items = self.get_items()
        )
        self.sni_service.register()

        self.update_tooltip()

    def initialize(self, plugin_config):
        """
        Initialize the tray icon by setting the config.
        """
        self.plugin_config = plugin_config
        self.allow_disabling = plugin_config['allow_disabling']

        self.update_menu()
        self.update_tooltip()

    def get_items(self):
        breaks_found = self.has_breaks()

        info_message = _('No Breaks Available')

        if breaks_found:
            if self.active:
                next_break = self.get_next_break_time()

                if next_break is not None:
                    (next_time, next_long_time, next_is_long) = next_break

                    if next_long_time:
                        if next_is_long:
                            info_message = _('Next long break at %s') % (next_long_time)
                        else:
                            info_message = _('Next breaks at %(short)s/%(long)s') % {
                                'short': next_time,
                                'long': next_long_time
                            }
                    else:
                        info_message = _('Next break at %s') % (next_time)
            else:
                if self.wakeup_time:
                    info_message = _('Disabled until %s') % utility.format_time(self.wakeup_time)
                else:
                    info_message = _('Disabled until restart')

        disable_items = []

        if self.allow_disabling:
            disable_option_dynamic_id = 13

            for disable_option in self.plugin_config['disable_options']:
                time_in_minutes = time_in_x = disable_option['time']

                # Validate time value
                if not isinstance(time_in_minutes, int) or time_in_minutes <= 0:
                    logging.error('Invalid time in disable option: ' + str(time_in_minutes))
                    continue
                time_unit = disable_option['unit'].lower()
                if time_unit == 'seconds' or time_unit == 'second':
                    time_in_minutes = int(time_in_minutes / 60)
                    label = self.context['locale'].ngettext(
                        'For %(num)d Second',
                        'For %(num)d Seconds',
                        time_in_x
                    ) % {'num': time_in_x}
                elif time_unit == 'minutes' or time_unit == 'minute':
                    time_in_minutes = int(time_in_minutes * 1)
                    label = self.context['locale'].ngettext(
                        'For %(num)d Minute',
                        'For %(num)d Minutes',
                        time_in_x
                    ) % {'num': time_in_x}
                elif time_unit == 'hours' or time_unit == 'hour':
                    time_in_minutes = int(time_in_minutes * 60)
                    label = self.context['locale'].ngettext(
                        'For %(num)d Hour',
                        'For %(num)d Hours',
                        time_in_x
                    ) % {'num': time_in_x}
                else:
                    # Invalid unit
                    logging.error('Invalid unit in disable option: ' + str(disable_option))
                    continue

                disable_items.append({
                    'id': disable_option_dynamic_id,
                    'label': label,
                    'callback': lambda: self.on_disable_clicked(time_in_minutes),
                })

                disable_option_dynamic_id += 1

            disable_items.append({
                'id': 12,
                'label': _('Until restart'),
                'callback': lambda: self.on_disable_clicked(-1),
            })

        return [
            {
                'id': 1,
                'label': info_message,
                'icon-name': "io.github.slgobinath.SafeEyes-timer",
                'enabled': breaks_found and self.active,
            },
            {
                'id': 2,
                'type': "separator",
            },
            {
                'id': 3,
                'label': _("Enable Safe Eyes"),
                'enabled': breaks_found and not self.active,
                'callback': self.on_enable_clicked,
                'hidden': not self.allow_disabling,
            },
            {
                'id': 4,
                'label': _("Disable Safe Eyes"),
                'enabled': breaks_found and self.active and not self.menu_locked,
                'children-display': 'submenu',
                'children': disable_items,
                'hidden': not self.allow_disabling,
            },
            {
                'id': 5,
                'label': _('Take a break now'),
                'enabled': breaks_found and self.active and not self.menu_locked,
                'children-display': 'submenu',
                'children': [
                    {
                        'id': 9,
                        'label': _('Any break'),
                        'callback': lambda: self.on_manual_break_clicked(None),
                    },
                    {
                        'id': 10,
                        'label': _('Short break'),
                        'callback': lambda: self.on_manual_break_clicked(BreakType.SHORT_BREAK),
                    },
                    {
                        'id': 11,
                        'label': _('Long break'),
                        'callback': lambda: self.on_manual_break_clicked(BreakType.LONG_BREAK),
                    },
                ]
            },
            {
                'id': 6,
                'label': _('Settings'),
                'enabled': not self.menu_locked,
                'callback': self.show_settings,
            },
            {
                'id': 7,
                'label': _('About'),
                'callback': self.show_about,
            },
            {
                'id': 8,
                'label': _('Quit'),
                'enabled': not self.menu_locked,
                'callback': self.quit_safe_eyes,
                'hidden': not self.allow_disabling,
            },
        ]

    def update_menu(self):
        self.sni_service.set_items(self.get_items())

    def update_tooltip(self):
        next_break = self.get_next_break_time()

        if next_break is not None and self.plugin_config.get('show_time_in_tray', False):
            (next_time, next_long_time, _next_is_long) = next_break

            if next_long_time and self.plugin_config.get('show_long_time_in_tray', False):
                description = next_long_time
            else:
                description = next_time
        else:
            description = ''

        self.sni_service.set_tooltip('Safe Eyes', description)

    def quit_safe_eyes(self):
        """
        Handle Quit menu action.
        This action terminates the application.
        """
        with self.lock:
            self.active = True
            # Notify all schedulers
            self.idle_condition.acquire()
            self.idle_condition.notify_all()
            self.idle_condition.release()
        self.quit()

    def show_settings(self):
        """
        Handle Settings menu action.
        This action shows the Settings dialog.
        """
        self.on_show_settings()

    def show_about(self):
        """
        Handle About menu action.
        This action shows the About dialog.
        """
        self.on_show_about()

    def next_break_time(self, dateTime):
        """
        Update the next break time to be displayed in the menu and optionally in the tray icon.
        """
        logging.info("Update next break information")
        self.date_time = dateTime
        self.update_menu()
        self.update_tooltip()

    def get_next_break_time(self):
        if not (self.has_breaks() and self.active and self.date_time):
            return None

        formatted_time = utility.format_time(self.get_break_time())
        long_time = self.get_break_time(BreakType.LONG_BREAK)

        if long_time:
            long_time = utility.format_time(long_time)
            if long_time == formatted_time:
                return (long_time, long_time, True)
            else:
                return (formatted_time, long_time, False)

        return (formatted_time, None, False)


    def on_manual_break_clicked(self, break_type):
        """
        Trigger a break manually.
        """
        self.take_break(break_type)

    def on_enable_clicked(self):
        """
        Handle 'Enable Safe Eyes' menu action.
        This action enables the application if it is currently disabled.
        """
        if not self.active:
            with self.lock:
                self.enable_ui()
                self.enable_safeeyes()
                # Notify all schedulers
                self.idle_condition.acquire()
                self.idle_condition.notify_all()
                self.idle_condition.release()

    def on_disable_clicked(self, time_to_wait):
        """
        Handle the menu actions of all the sub menus of 'Disable Safe Eyes'.
        This action disables the application if it is currently active.
        """
        if self.active:
            self.disable_ui()

            if time_to_wait <= 0:
                info = _('Disabled until restart')
                self.disable_safeeyes(info)
                self.wakeup_time = None
            else:
                self.wakeup_time = datetime.datetime.now() + datetime.timedelta(minutes=time_to_wait)
                info = _('Disabled until %s') % utility.format_time(self.wakeup_time)
                self.disable_safeeyes(info)
                utility.start_thread(self.__schedule_resume, time_minutes=time_to_wait)
            self.update_menu()

    def lock_menu(self):
        """
        This method is called by the core to prevent user from disabling Safe Eyes after the notification.
        """
        if self.active:
            self.menu_locked = True
            self.update_menu()

    def unlock_menu(self):
        """
        This method is called by the core to activate the menu after the the break.
        """
        if self.active:
            self.menu_locked = False
            self.update_menu()

    def disable_ui(self):
        """
        Change the UI to disabled state.
        """
        if self.active:
            logging.info('Disable Safe Eyes')
            self.active = False

            self.sni_service.set_icon("io.github.slgobinath.SafeEyes-disabled")
            self.update_menu()

    def enable_ui(self):
        """
        Change the UI to enabled state.
        """
        if not self.active:
            logging.info('Enable Safe Eyes')
            self.active = True

            self.sni_service.set_icon("io.github.slgobinath.SafeEyes-enabled")
            self.update_menu()

    def __schedule_resume(self, time_minutes):
        """
        Schedule a local timer to enable Safe Eyes after the given timeout.
        """
        self.idle_condition.acquire()
        self.idle_condition.wait(time_minutes * 60)    # Convert to seconds
        self.idle_condition.release()

        with self.lock:
            if not self.active:
                utility.execute_main_thread(self.on_enable_clicked)

    def start_animation(self):
        if not self.active or not self.animate:
            return
        utility.execute_main_thread(lambda: self.sni_service.set_icon("io.github.slgobinath.SafeEyes-disabled"))
        time.sleep(0.5)
        utility.execute_main_thread(lambda: self.sni_service.set_icon("io.github.slgobinath.SafeEyes-enabled"))
        if self.animate and self.active:
            time.sleep(0.5)
            if self.animate and self.active:
                utility.start_thread(self.start_animation)

    def stop_animation(self):
        self.animate = False
        if self.active:
            utility.execute_main_thread(lambda: self.sni_service.set_icon("io.github.slgobinath.SafeEyes-enabled"))
        else:
            utility.execute_main_thread(lambda: self.sni_service.set_icon("io.github.slgobinath.SafeEyes-disabled"))

def init(ctx, safeeyes_cfg, plugin_config):
    """
    Initialize the tray icon.
    """
    global context
    global tray_icon
    global safeeyes_config
    logging.debug('Initialize Tray Icon plugin')
    context = ctx
    safeeyes_config = safeeyes_cfg
    if not tray_icon:
        tray_icon = TrayIcon(context, plugin_config)
    else:
        tray_icon.initialize(plugin_config)


def update_next_break(break_obj, next_break_time):
    """
    Update the next break time.
    """
    tray_icon.next_break_time(next_break_time)


def on_pre_break(break_obj):
    """
    Disable the menu if strict_break is enabled
    """
    if safeeyes_config.get('strict_break'):
        tray_icon.lock_menu()
    tray_icon.animate = True
    tray_icon.start_animation()


def on_start_break(break_obj):
    tray_icon.stop_animation()


def on_stop_break():
    tray_icon.unlock_menu()


def on_start():
    """
    Enable the tray icon.
    """
    tray_icon.enable_ui()


def on_stop():
    """
    Disable the tray icon.
    """
    tray_icon.disable_ui()



#!/usr/bin/env python
# Safe Eyes is a utility to remind you to take break frequently
# to protect your eyes from eye strain.

# Copyright (C) 2017  Gobinath

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
PluginManager loads all enabled plugins and call their lifecycle methods.
A plugin must have the following directory structure:
<plugin-id>
    |- config.json
    |- plugin.py
    |- icon.png (Optional)

The plugin.py can have following methods but all are optional:
 - description()
    If a custom description has to be displayed, use this function
 - init(context, safeeyes_config, plugin_config)
    Initialize the plugin. Will be called after loading and after every changes in configuration
 - on_start()
    Executes when Safe Eyes is enabled
 - on_stop()
    Executes when Safe Eyes is disabled
 - on_exit()
    Executes before Safe Eyes exits
 - on_pre_break(break_obj)
    Executes at the start of the prepare time for a break
 - on_start_break(break_obj)
    Executes when a break starts
 - on_stop_break()
    Executes when a break stops
 - on_countdown(countdown, seconds)
    Executes every second throughout a break
 - update_next_break(break_obj, break_time)
    Executes when the next break changes
 - enable()
    Executes once the plugin.py is loaded as a module
 - disable()
    Executes if the plugin is disabled at the runtime by the user
"""

import importlib
import logging
import os
import sys

from safeeyes import utility
from safeeyes.model import RequiredPluginException

sys.path.append(os.path.abspath(utility.SYSTEM_PLUGINS_DIR))
sys.path.append(os.path.abspath(utility.USER_PLUGINS_DIR))

HORIZONTAL_LINE_LENGTH = 64


class PluginManager:
    """
    Imports the Safe Eyes plugins and calls the methods defined in those plugins.
    """

    def __init__(self):
        logging.info('Load all the plugins')
        self.__plugins = {}
        self.__plugins_on_init = []
        self.__plugins_on_start = []
        self.__plugins_on_stop = []
        self.__plugins_on_exit = []
        self.__plugins_on_pre_break = []
        self.__plugins_on_start_break = []
        self.__plugins_on_stop_break = []
        self.__plugins_on_countdown = []
        self.__plugins_update_next_break = []
        self.__widget_plugins = []
        self.__tray_actions_plugins = []
        self.last_break = None
        self.horizontal_line = '─' * HORIZONTAL_LINE_LENGTH

    def init(self, context, config):
        """
        Initialize all the plugins with init(context, safe_eyes_config, plugin_config) function.
        """
        # Load the plugins
        for plugin in config.get('plugins'):
            try:
                self.__load_plugin(plugin)
            except RequiredPluginException as e:
                raise e
            except BaseException as e:
                traceback_wanted = logging.getLogger().getEffectiveLevel() == logging.DEBUG
                if traceback_wanted:
                    import traceback
                    traceback.print_exc()
                logging.error('Error in loading the plugin %s: %s', plugin['id'], e)
                continue
        # Initialize the plugins
        for plugin in self.__plugins_on_init:
            plugin['module'].init(context, config, plugin['config'])
        return True

    def start(self):
        """
        Execute the on_start() function of plugins.
        """
        for plugin in self.__plugins_on_start:
            plugin['module'].on_start()
        return True

    def stop(self):
        """
        Execute the on_stop() function of plugins.
        """
        for plugin in self.__plugins_on_stop:
            plugin['module'].on_stop()
        return True

    def exit(self):
        """
        Execute the on_exit() function of plugins.
        """
        for plugin in self.__plugins_on_exit:
            plugin['module'].on_exit()
        return True

    def pre_break(self, break_obj):
        """
        Execute the on_pre_break(break_obj) function of plugins.
        """
        for plugin in self.__plugins_on_pre_break:
            if break_obj.plugin_enabled(plugin['id'], plugin['enabled']):
                if plugin['module'].on_pre_break(break_obj):
                    return False
        return True

    def start_break(self, break_obj):
        """
        Execute the start_break(break_obj) function of plugins.
        """
        self.last_break = break_obj
        for plugin in self.__plugins_on_start_break:
            if break_obj.plugin_enabled(plugin['id'], plugin['enabled']):
                if plugin['module'].on_start_break(break_obj):
                    return False

        return True

    def stop_break(self):
        """
        Execute the stop_break() function of plugins.
        """
        for plugin in self.__plugins_on_stop_break:
            if self.last_break.plugin_enabled(plugin['id'], plugin['enabled']):
                plugin['module'].on_stop_break()

    def countdown(self, countdown, seconds):
        """
        Execute the on_countdown(countdown, seconds) function of plugins.
        """
        for plugin in self.__plugins_on_countdown:
            if self.last_break.plugin_enabled(plugin['id'], plugin['enabled']):
                plugin['module'].on_countdown(countdown, seconds)

    def update_next_break(self, break_obj, break_time):
        """
        Execute the update_next_break(break_time) function of plugins.
        """
        for plugin in self.__plugins_update_next_break:
            plugin['module'].update_next_break(break_obj, break_time)
        return True

    def get_break_screen_widgets(self, break_obj):
        """
        Return the HTML widget generated by the plugins.
        The widget is generated by calling the get_widget_title and get_widget_content functions of plugins.
        """
        widget = ''
        for plugin in self.__widget_plugins:
            if break_obj.plugin_enabled(plugin['id'], plugin['enabled']):
                try:
                    title = plugin['module'].get_widget_title(break_obj).upper().strip()
                    if title == '':
                        continue
                    content = plugin['module'].get_widget_content(break_obj)
                    if content == '':
                        continue
                    widget += '<b>{}</b>\n{}\n{}\n\n\n'.format(title, self.horizontal_line, content)
                except BaseException:
                    continue
        return widget.strip()

    def get_break_screen_tray_actions(self, break_obj):
        """
        Return Tray Actions.
        """
        actions = []
        for plugin in self.__tray_actions_plugins:
            if break_obj.plugin_enabled(plugin['id'], plugin['enabled']):
                action = plugin['module'].get_tray_action(break_obj)
                if action:
                    actions.append(action)

        return actions

    def __load_plugin(self, plugin):
        """
        Load the given plugin.
        """
        plugin_enabled = plugin['enabled']
        if plugin['id'] in self.__plugins and not plugin_enabled:
            # A disabled plugin but that was loaded earlier
            plugin_obj = self.__plugins[plugin['id']]
            if plugin_obj['enabled']:
                # Previously enabled but now disabled
                plugin_obj['enabled'] = False
                utility.remove_if_exists(self.__plugins_on_start, plugin_obj)
                utility.remove_if_exists(self.__plugins_on_stop, plugin_obj)
                utility.remove_if_exists(self.__plugins_on_exit, plugin_obj)
                utility.remove_if_exists(self.__plugins_update_next_break, plugin_obj)
                # Call the plugin.disable method if available
                if utility.has_method(plugin_obj['module'], 'disable'):
                    plugin_obj['module'].disable()
                logging.info("Successfully unloaded the plugin '%s'", plugin['id'])

            if not plugin_obj['break_override_allowed']:
                # Remaining methods also should be removed
                utility.remove_if_exists(self.__plugins_on_init, plugin_obj)
                utility.remove_if_exists(self.__plugins_on_pre_break, plugin_obj)
                utility.remove_if_exists(self.__plugins_on_start_break, plugin_obj)
                utility.remove_if_exists(self.__plugins_on_stop_break, plugin_obj)
                utility.remove_if_exists(self.__plugins_on_countdown, plugin_obj)
                utility.remove_if_exists(self.__widget_plugins, plugin_obj)
                utility.remove_if_exists(self.__tray_actions_plugins, plugin_obj)
                del self.__plugins[plugin['id']]
            return

        # Look for plugin.py
        if os.path.isfile(os.path.join(utility.SYSTEM_PLUGINS_DIR, plugin['id'], 'plugin.py')):
            plugin_dir = utility.SYSTEM_PLUGINS_DIR
        elif os.path.isfile(os.path.join(utility.USER_PLUGINS_DIR, plugin['id'], 'plugin.py')):
            plugin_dir = utility.USER_PLUGINS_DIR
        else:
            logging.error('plugin.py not found for the plugin: %s', plugin['id'])
            return
        # Look for config.json
        plugin_path = os.path.join(plugin_dir, plugin['id'])
        plugin_config_path = os.path.join(plugin_path, 'config.json')
        if not os.path.isfile(plugin_config_path):
            logging.error('config.json not found for the plugin: %s', plugin['id'])
            return
        plugin_config = utility.load_json(plugin_config_path)
        if plugin_config is None:
            return

        if plugin_enabled or plugin_config.get('break_override_allowed', False):
            if plugin['id'] in self.__plugins:
                # The plugin is already enabled or partially loaded due to break_override_allowed

                # Validate the dependencies again
                if utility.check_plugin_dependencies(plugin['id'], plugin_config, plugin.get('settings', {}), plugin_path):
                    plugin_obj['enabled'] = False
                    utility.remove_if_exists(self.__plugins_on_start, plugin_obj)
                    utility.remove_if_exists(self.__plugins_on_stop, plugin_obj)
                    utility.remove_if_exists(self.__plugins_on_exit, plugin_obj)
                    utility.remove_if_exists(self.__plugins_update_next_break, plugin_obj)
                    utility.remove_if_exists(self.__plugins_on_init, plugin_obj)
                    utility.remove_if_exists(self.__plugins_on_pre_break, plugin_obj)
                    utility.remove_if_exists(self.__plugins_on_start_break, plugin_obj)
                    utility.remove_if_exists(self.__plugins_on_stop_break, plugin_obj)
                    utility.remove_if_exists(self.__plugins_on_countdown, plugin_obj)
                    utility.remove_if_exists(self.__widget_plugins, plugin_obj)
                    utility.remove_if_exists(self.__tray_actions_plugins, plugin_obj)
                    del self.__plugins[plugin['id']]

                # Use the existing plugin object
                plugin_obj = self.__plugins[plugin['id']]

                # Update the config
                plugin_obj['config'] = dict(plugin.get('settings', {}))
                plugin_obj['config']['path'] = os.path.join(plugin_dir, plugin['id'])

                if plugin_obj['enabled']:
                    # Already loaded completely
                    return
                # Plugin was partially loaded due to break_override_allowed
                if plugin_enabled:
                    # Load the rest of the methods
                    plugin_obj['enabled'] = True
                    module = plugin_obj['module']
                    self.__init_plugin(module, plugin_obj)
            else:
                # This is the first time to load the plugin
                # Check for dependencies
                message = utility.check_plugin_dependencies(plugin['id'], plugin_config, plugin.get('settings', {}), plugin_path)
                if message:
                    if plugin_config.get('required_plugin', False):
                        raise RequiredPluginException(
                            plugin['id'],
                            plugin_config['meta']['name'],
                            message
                        )
                    return

                # Load the plugin module
                module = importlib.import_module((plugin['id'] + '.plugin'))
                logging.info("Successfully loaded %s", str(module))
                plugin_obj = {'id': plugin['id'], 'module': module, 'config': dict(plugin.get(
                    'settings', {})), 'enabled': plugin_enabled,
                              'break_override_allowed': plugin_config.get('break_override_allowed', False)}
                # Inject the plugin directory into the config
                plugin_obj['config']['path'] = os.path.join(plugin_dir, plugin['id'])
                self.__plugins[plugin['id']] = plugin_obj
                if utility.has_method(module, 'enable'):
                    module.enable()
                if plugin_enabled:
                    self.__init_plugin(module, plugin_obj)
                if utility.has_method(module, 'init', 3):
                    self.__plugins_on_init.append(plugin_obj)
                if utility.has_method(module, 'on_pre_break', 1):
                    self.__plugins_on_pre_break.append(plugin_obj)
                if utility.has_method(module, 'on_start_break', 1):
                    self.__plugins_on_start_break.append(plugin_obj)
                if utility.has_method(module, 'on_stop_break', 0):
                    self.__plugins_on_stop_break.append(plugin_obj)
                if utility.has_method(module, 'on_countdown', 2):
                    self.__plugins_on_countdown.append(plugin_obj)
                if utility.has_method(module, 'get_widget_title', 1) and utility.has_method(module,
                                                                                            'get_widget_content', 1):
                    self.__widget_plugins.append(plugin_obj)
                if utility.has_method(module, 'get_tray_action', 1):
                    self.__tray_actions_plugins.append(plugin_obj)

    def __init_plugin(self, module, plugin_obj):
        """
        Collect mandatory methods from the plugin and add them to the life cycle methods list.
        """
        if utility.has_method(module, 'on_start'):
            self.__plugins_on_start.append(plugin_obj)
        if utility.has_method(module, 'on_stop'):
            self.__plugins_on_stop.append(plugin_obj)
        if utility.has_method(module, 'on_exit'):
            self.__plugins_on_exit.append(plugin_obj)
        if utility.has_method(module, 'update_next_break', 2):
            self.__plugins_update_next_break.append(plugin_obj)



#!/usr/bin/env python
# Safe Eyes is a utility to remind you to take break frequently
# to protect your eyes from eye strain.

# Copyright (C) 2016  Gobinath

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
This module creates the RequiredPluginDialog which shows the error for a required plugin.
"""

import os

from safeeyes import utility
from safeeyes.model import PluginDependency

REQUIRED_PLUGIN_DIALOG_GLADE = os.path.join(utility.BIN_DIRECTORY, "glade/required_plugin_dialog.glade")


class RequiredPluginDialog:
    """
    RequiredPluginDialog shows an error when a plugin has required dependencies.
    """

    def __init__(self, plugin_id, plugin_name, message, on_quit, on_disable_plugin):
        self.on_quit = on_quit
        self.on_disable_plugin = on_disable_plugin

        builder = utility.create_gtk_builder(REQUIRED_PLUGIN_DIALOG_GLADE)
        self.window = builder.get_object('window_required_plugin')

        self.window.connect("delete-event", self.on_window_delete)
        builder.get_object('btn_close').connect('clicked', self.on_close_clicked)
        builder.get_object('btn_disable_plugin').connect('clicked', self.on_disable_plugin_clicked)

        builder.get_object('lbl_header').set_label(_("The required plugin '%s' is missing dependencies!") % _(plugin_name))

        builder.get_object('lbl_main').set_label(_("Please install the dependencies or disable the plugin. To hide this message, you can also deactivate the plugin in the settings."))

        builder.get_object('btn_close').set_label(_("Quit"))
        builder.get_object('btn_disable_plugin').set_label(_("Disable plugin temporarily"))

        if isinstance(message, PluginDependency):
            builder.get_object('lbl_message').set_label(_(message.message))
            btn_extra_link = builder.get_object('btn_extra_link')
            btn_extra_link.set_label(_("Click here for more information"))
            btn_extra_link.set_uri(message.link)
            btn_extra_link.set_visible(True)
        else:
            builder.get_object('lbl_message').set_label(_(message))

    def show(self):
        """
        Show the dialog.
        """
        self.window.show_all()

    def on_window_delete(self, *args):
        """
        Window close event handler.
        """
        self.window.destroy()
        self.on_quit()

    def on_close_clicked(self, *args):
        self.window.destroy()
        self.on_quit()

    def on_disable_plugin_clicked(self, *args):
        self.window.destroy()
        self.on_disable_plugin()



#!/usr/bin/env python
# Safe Eyes is a utility to remind you to take break frequently
# to protect your eyes from eye strain.

# Copyright (C) 2017  Gobinath

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
Show health statistics on the break screen.
"""

import croniter
import datetime
import logging

context = None
session = None
statistics_reset_cron = None
default_statistics_reset_cron = '0 0 * * *'  # Every midnight
next_reset_time = None
start_time = None

def init(ctx, safeeyes_config, plugin_config):
    """
    Initialize the plugin.
    """
    global context
    global session
    global statistics_reset_cron

    logging.debug('Initialize Health Stats plugin')
    context = ctx
    statistics_reset_cron = plugin_config.get('statistics_reset_cron', default_statistics_reset_cron)

    if session is None:
        # Read the session
        defaults = {
            'breaks': 0,
            'skipped_breaks': 0,
            'screen_time': 0,
            'total_breaks': 0,
            'total_skipped_breaks': 0,
            'total_screen_time': 0,
            'total_resets': 0,
        }

        session = context['session']['plugin'].get('healthstats', {}) | defaults
        if 'no_of_breaks' in session:
            # Ignore old format session.
            session = defaults
        context['session']['plugin']['healthstats'] = session

    _get_next_reset_time()


def on_stop_break():
    # Check if break was skipped.
    global session
    if context['skipped']:
        session['skipped_breaks'] += 1

    # Screen time is starting again.
    on_start()


def on_start_break(break_obj):
    global session
    session['breaks'] += 1

    # Screen time has stopped.
    on_stop()


def on_stop():
    global start_time
    _reset_stats()
    if start_time:
        screen_time = datetime.datetime.now() - start_time
        session['screen_time'] += round(screen_time.total_seconds())
        start_time = None


def get_widget_title(break_obj):
    """
    Return the widget title.
    """
    return _('Health Statistics')


def _reset_stats():
    global session

    # Check if the reset time has passed
    if next_reset_time and datetime.datetime.now() >= next_reset_time:
        logging.info("Resetting the health statistics")

        # Update the next_reset_time
        _get_next_reset_time()

        # Reset statistics
        session['total_breaks'] += session['breaks']
        session['total_skipped_breaks'] += session['skipped_breaks']
        session['total_screen_time'] += session['screen_time']
        session['total_resets'] += 1
        session['breaks'] = 0
        session['skipped_breaks'] = 0
        session['screen_time'] = 0


def get_widget_content(break_obj):
    """
    Return the statistics.
    """
    global next_reset_time
    resets = session['total_resets']
    if session['screen_time'] > 21600 or (session['breaks'] and session['skipped_breaks'] / session['breaks']) >= 0.2:
        # Unhealthy behavior -> Red broken heart
        heart = '💔️'
    else:
        # Healthy behavior -> Green heart
        heart = '💚'

    content = [
      heart,
      f"BREAKS: {session['breaks']}",
      f"SKIPPED: {session['skipped_breaks']}",
      f"SCREEN TIME: {_format_interval(session['screen_time'])}",
    ]

    if resets:
        content[1] += f" [{round(session['total_breaks'] / resets, 1)}]"
        content[2] += f" [{round(session['total_skipped_breaks'] / resets, 1)}]"
        content[3] += f" [{_format_interval(session['total_screen_time'] / resets)}]"

    content = "\t".join(content)
    if resets:
        content += f"\n\t[] = average of {resets} reset(s)"
    if next_reset_time is None:
        content += f"\n\tSettings error in statistics reset interval: {statistics_reset_cron}"
    return content


def on_start():
    """
    Track the start time.
    """
    global start_time
    _reset_stats()
    start_time = datetime.datetime.now()


def _get_next_reset_time():
    global next_reset_time
    global session

    try:
        cron = croniter.croniter(statistics_reset_cron, datetime.datetime.now())
        next_reset_time = cron.get_next(datetime.datetime)
        session['next_reset_time'] = next_reset_time.strftime("%Y-%m-%d %H:%M:%S")
        logging.debug("Health stats will be reset at " + session['next_reset_time'])
    except:
        logging.error("Error in statistics reset expression: " + statistics_reset_cron)
        next_reset_time = None


def _format_interval(seconds):
    screen_time = round(seconds / 60)
    hours, minutes = divmod(screen_time, 60)
    return '{:02d}:{:02d}'.format(hours, minutes)



#!/usr/bin/env python
# Safe Eyes is a utility to remind you to take break frequently
# to protect your eyes from eye strain.

# Copyright (C) 2019  Gobinath

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
Media Control plugin lets users to pause currently playing media player from the break screen.
"""

import logging
import os
import re
import gi
from safeeyes.model import TrayAction
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, Gio

tray_icon_path = None


def __active_players():
    """
    List of all media players which are playing now.
    """
    players = []

    dbus_proxy = Gio.DBusProxy.new_for_bus_sync(
        bus_type=Gio.BusType.SESSION,
        flags=Gio.DBusProxyFlags.DO_NOT_LOAD_PROPERTIES,
        info=None,
        name='org.freedesktop.DBus',
        object_path='/org/freedesktop/DBus',
        interface_name='org.freedesktop.DBus',
        cancellable=None,
    )

    for service in dbus_proxy.ListNames():
        if re.match('org.mpris.MediaPlayer2.', service):
            player = Gio.DBusProxy.new_for_bus_sync(
                bus_type=Gio.BusType.SESSION,
                flags=Gio.DBusProxyFlags.NONE,
                info=None,
                name=service,
                object_path='/org/mpris/MediaPlayer2',
                interface_name='org.mpris.MediaPlayer2.Player',
                cancellable=None,
            )

            status = player.get_cached_property('PlaybackStatus').unpack().lower()

            if status == "playing":
                players.append(player)
    return players


def __pause_players(players):
    """
    Pause all playing media players using dbus.
    """
    for player in players:
        player.Pause()


def init(ctx, safeeyes_config, plugin_config):
    """
    Initialize the screensaver plugin.
    """
    global tray_icon_path
    tray_icon_path = os.path.join(plugin_config['path'], "resource/pause.png")


def get_tray_action(break_obj):
    """
    Return TrayAction only if there is a media player currently playing.
    """
    players = __active_players()
    if players:
        return TrayAction.build("Pause media",
                                tray_icon_path,
                                Gtk.STOCK_MEDIA_PAUSE,
                                lambda: __pause_players(players))
